
/**
 * @file   linux/at/at_linux_pci.c
 *
 * @author Alazar Technologies Inc
 *
 * @copyright Copyright (c) 2020 Alazar Technologies Inc. All Rights
 * Reserved.  Unpublished - rights reserved under the Copyright laws
 * of the United States And Canada.
 * This product contains confidential information and trade secrets
 * of Alazar Technologies Inc. Use, disclosure, or reproduction is
 * prohibited without the prior express written permission of Alazar
 * Technologies Inc
 */

// This file is included into the dkms package.

// Includes
/////////////////////////////////////////////////////////////////////////////

// ===== Linux kernel =======================================================
#include <linux/pci.h>
#include <linux/version.h>

// ===== linux/at ===========================================================
#include <at_assert.h>
#include <at_main.h>

#include <at_linux_pci.h>

// ===== src/ats... =========================================================
#include <DeviceId.h>

// Functions
/////////////////////////////////////////////////////////////////////////////

#undef AT_DEBUG
#ifdef AT_DEBUG
#define DBG_PRINT(...) printk(KERN_DEBUG __VA_ARGS__)
#else
#define DBG_PRINT(...)
#endif

// aPciDev  [---;RW-]
// aDmaAddr [---;-W-]
void *at_pci_alloc_consistent(struct pci_dev *aPciDev, size_t aSize_byte, dma_addr_t *aDmaAddr)
{
    void *l_result;

    at_assert(NULL != aPciDev);
    at_assert(0 < aSize_byte);
    at_assert(NULL != aDmaAddr);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0)
    l_result = dma_alloc_coherent(&aPciDev->dev, aSize_byte, aDmaAddr, GFP_ATOMIC);
#else
    l_result = pci_alloc_consistent(aPciDev, aSize_byte, aDmaAddr);
#endif
    if (NULL == l_result)
    {
        printk(KERN_ERR "at_pci_alloc_consistent - pci_alloc_consistent( , %lu bytes,  ) failed\n",
               aSize_byte);
    }

    return l_result;
}

// aPciDev  [---;RW-]
bool at_pci_alloc_irq_vectors_MSI(struct pci_dev *aPciDev, unsigned int aCount)
{
    DBG_PRINT("%s(%p, %u)\n", __FUNCTION__, aPciDev, aCount);
    at_assert(NULL != aPciDev);
    at_assert(0 < aCount);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0)
    return (pci_alloc_irq_vectors(aPciDev, aCount, aCount, PCI_IRQ_MSI) == aCount);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3, 16, 0)
    return (pci_enable_msi_range(aPciDev, aCount, aCount) == 0);
#else
    return (pci_enable_msi_block(aPciDev, aCount) == 0);
#endif
}

// aPciDev  [---;RW-]
// aEntries [---;RW-]
bool at_pci_alloc_irq_vectors_MSIX(struct pci_dev *aPciDev,
                                   unsigned int aCount,
                                   struct msix_entry *aEntries)
{
    int ret;

    DBG_PRINT("%s(%p, %u, %p)\n", __FUNCTION__, aPciDev, aCount, aEntries);
    at_assert(NULL != aPciDev);
    at_assert(0 < aCount);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0)
    ret = pci_enable_msix_exact(aPciDev, aEntries, aCount);
    DBG_PRINT("%s(), pci_enable_msix_exact() return %d\n", __FUNCTION__, ret);
#else
    ret = pci_enable_msix(aPciDev, aEntries, aCount);
    DBG_PRINT("%s(), pci_enable_msix() return %d\n", __FUNCTION__, ret);
#endif

    if (ret == 0) return true;

    printk(KERN_ERR "at_pci_alloc_irq_vectors_MSIX - pci_enable_msix_exact( , , %u ) failed\n",
           aCount);
    return false;
}

// aPciDev [---;RW-]
void at_pci_disable_device(struct pci_dev *aPciDev)
{
    DBG_PRINT("%s(%p)\n", __FUNCTION__, aPciDev);
    at_assert(NULL != aPciDev);
    pci_disable_device(aPciDev);
}

// aPciDev [---;RW-]
void at_pci_disable_msi(struct pci_dev *aPciDev)
{
    DBG_PRINT("%s(%p)\n", __FUNCTION__, aPciDev);
    at_assert(NULL != aPciDev);
    pci_disable_msi(aPciDev);
}

// aPciDev [---;RW-]
void at_pci_disable_msix(struct pci_dev *aPciDev)
{
    DBG_PRINT("%s(%p)\n", __FUNCTION__, aPciDev);
    at_assert(NULL != aPciDev);
    pci_disable_msix(aPciDev);
}

// aPciDev [---;RW-]
int at_pci_enable_device(struct pci_dev *aPciDev)
{
    DBG_PRINT("%s(%p)\n", __FUNCTION__, aPciDev);
    at_assert(NULL != aPciDev);
    return pci_enable_device(aPciDev);
}

// aPciDev [---;RW-]
int at_pci_enable_msi(struct pci_dev *aPciDev)
{
    DBG_PRINT("%s(%p)\n", __FUNCTION__, aPciDev);
    at_assert(NULL != aPciDev);
    return pci_enable_msi(aPciDev);
}

// aPciDev [---;R--]
int at_pci_find_capability(struct pci_dev *aPciDev, int aCap)
{
    DBG_PRINT("%s(%p)\n", __FUNCTION__, aPciDev);
    at_assert(NULL != aPciDev);
    return pci_find_capability(aPciDev, aCap);
}

// aPciDev   [---;RW-]
// aVirtAddr [D--;---]
void at_pci_free_consistent(struct pci_dev *aPciDev,
                            uint64_t aSize_byte,
                            void *aVirtAddr,
                            uint64_t aHandle)
{
    DBG_PRINT("%s(%p, %llu, %p, %llx)\n", __FUNCTION__, aPciDev, aSize_byte, aVirtAddr, aHandle);
    at_assert(NULL != aPciDev);
    at_assert(0 < aSize_byte);
    at_assert(NULL != aVirtAddr);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0)
    dma_free_coherent(&aPciDev->dev, aSize_byte, aVirtAddr, aHandle);
#else
    pci_free_consistent(aPciDev, aSize_byte, aVirtAddr, aHandle);
#endif
}

// aPciDev [---;R--]
uint8_t at_pci_get_bus_number(struct pci_dev *aPciDev)
{
    at_assert(NULL != aPciDev);
    return aPciDev->bus->number;
}
uint32_t at_pci_get_class(struct pci_dev *aPciDev)
{
    at_assert(NULL != aPciDev);
    return aPciDev->class;
}
struct device *at_pci_get_device(struct pci_dev *aPciDev)
{
    at_assert(NULL != aPciDev);
    return &aPciDev->dev;
}
uint16_t at_pci_get_device_id(struct pci_dev *aPciDev)
{
    at_assert(NULL != aPciDev);
    return aPciDev->device;
}

// aPciDev [---;R--]
unsigned int at_pci_get_irq(struct pci_dev *aPciDev)
{
    at_assert(NULL != aPciDev);
    return aPciDev->irq;
}

uint8_t at_pci_get_revision(struct pci_dev *aPciDev)
{
    at_assert(NULL != aPciDev);
    return aPciDev->revision;
}
uint8_t at_pci_get_slot(struct pci_dev *aPciDev)
{
    at_assert(NULL != aPciDev);
    return PCI_SLOT(aPciDev->devfn);
}
uint16_t at_pci_get_vendor_id(struct pci_dev *aPciDev)
{
    at_assert(NULL != aPciDev);
    return aPciDev->vendor;
}

// aPciDev [---;R--]
int at_pci_irq_vector(struct pci_dev *aPciDev, unsigned int aIndex)
{
    DBG_PRINT("%s(%p, %u)\n", __FUNCTION__, aPciDev, aIndex);
    at_assert(NULL != aPciDev);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0)
    return pci_irq_vector(aPciDev, aIndex);
#else
    return (aPciDev->irq + aIndex);
#endif
}

// aPciDev [---;R--]
bool at_pci_is_irq_zero(struct pci_dev *aPciDev)
{
    DBG_PRINT("%s(%p)\n", __FUNCTION__, aPciDev);
    at_assert(NULL != aPciDev);
    return (0 == aPciDev->irq);
}

// aPciDev [---;R--]
void *at_pci_ioremap(struct pci_dev *aPciDev, unsigned int aIndex)
{
    DBG_PRINT("%s(%p, %u)\n", __FUNCTION__, aPciDev, aIndex);
    at_assert(NULL != aPciDev);
    return ioremap(pci_resource_start(aPciDev, aIndex), pci_resource_len(aPciDev, aIndex));
}

// aVirtAddr [D--;---]
void at_pci_iounmap(void *aVirtAddr)
{
    DBG_PRINT("%s(%p)\n", __FUNCTION__, aVirtAddr);
    at_assert(NULL != aVirtAddr);
    iounmap(aVirtAddr);
}

// aPciDev [---;RW-]
// aSgList [---;RW-]
uint32_t
at_pci_map_sg(struct pci_dev *aPciDev, void *aSgList, uint32_t aSgCount, bool aReadFromCard)
{
    DBG_PRINT("%s(%p, %p, %u, %u)\n", __FUNCTION__, aPciDev, aSgList, aSgCount, aReadFromCard);
    at_assert(NULL != aPciDev);
    at_assert(NULL != aSgList);
    at_assert(0 < aSgCount);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0)
    return dma_map_sg(&aPciDev->dev, aSgList, aSgCount,
                      aReadFromCard ? DMA_FROM_DEVICE : DMA_TO_DEVICE);
#else
    return pci_map_sg(aPciDev, aSgList, aSgCount,
                      aReadFromCard ? PCI_DMA_FROMDEVICE : PCI_DMA_TODEVICE);
#endif
}

// aPciDev [---;R--]
// aValue  [---;-W-]
int at_pci_read_config_byte(struct pci_dev *aPciDev, int aWhere, uint8_t *aValue)
{
    at_assert(NULL != aPciDev);
    at_assert(NULL != aValue);
    return pci_read_config_byte(aPciDev, aWhere, aValue);
}

// aPciDev [---;R--]
// aValue  [---;-W-]
int at_pci_read_config_bytes(struct pci_dev *aPciDev,
                             int aWhere,
                             uint8_t *aValues,
                             unsigned int aCount)
{
    unsigned int i;
    int lResult = 0;

    at_assert(NULL != aPciDev);
    at_assert(NULL != aValues);
    at_assert(0 < aCount);

    for (i = 0; (0 == lResult) && (i < aCount); i++)
    {
        lResult = at_pci_read_config_byte(aPciDev, aWhere + i, aValues + i);
    }

    return lResult;
}

// aPciDev [---;R--]
// aValue  [---;-W-]
int at_pci_read_config_dword(struct pci_dev *aPciDev, int aWhere, uint32_t *aValue)
{
    at_assert(NULL != aPciDev);
    at_assert(NULL != aValue);
    return pci_read_config_dword(aPciDev, aWhere, aValue);
}

// aPciDev [---;R--]
// aValue  [---;-W-]
int at_pci_read_config_word(struct pci_dev *aPciDev, int aWhere, uint16_t *aValue)
{
    at_assert(NULL != aPciDev);
    at_assert(NULL != aValue);
    return pci_read_config_word(aPciDev, aWhere, aValue);
}

// aPciDev [---;RW-]
void at_pci_release_regions(struct pci_dev *aPciDev)
{
    at_assert(NULL != aPciDev);
    pci_release_regions(aPciDev);
}

// aPciDev [---;RW-]
int at_pci_request_regions(struct pci_dev *aPciDev)
{
    at_assert(NULL != aPciDev);
    return pci_request_regions(aPciDev, DRIVER_NAME);
}

// aPciDev [---;R--]
bool at_pci_resource_io(struct pci_dev *aPciDev, unsigned int aIndex)
{
    at_assert(NULL != aPciDev);
    return (0 != (pci_resource_flags(aPciDev, aIndex) & IORESOURCE_IO));
}

// aPciDev [---;R--]
uint64_t at_pci_resource_len(struct pci_dev *aPciDev, unsigned int aIndex)
{
    at_assert(NULL != aPciDev);
    return pci_resource_len(aPciDev, aIndex);
}

// aPciDev [---;R--]
bool at_pci_resource_mem(struct pci_dev *aPciDev, unsigned int aIndex)
{
    at_assert(NULL != aPciDev);
    return (0 != (pci_resource_flags(aPciDev, aIndex) & IORESOURCE_MEM));
}

// aPciDev [---;R--]
uint64_t at_pci_resource_start(struct pci_dev *aPciDev, unsigned int aIndex)
{
    at_assert(NULL != aPciDev);
    return pci_resource_start(aPciDev, aIndex);
}

// aPciDev [---;RW-]
int at_pci_set_dma_mask(struct pci_dev *aPciDev, unsigned int aBits)
{
    u64 lBits;
    int lResult;

    DBG_PRINT("%s(%p, %u)\n", __FUNCTION__, aPciDev, aBits);
    at_assert(NULL != aPciDev);
    at_assert(0 < aBits);

    lBits = DMA_BIT_MASK(aBits);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0)
    lResult = dma_set_mask(&aPciDev->dev, lBits);
    if (0 == lResult) { lResult = dma_set_coherent_mask(&aPciDev->dev, lBits); }
#else
    lResult = pci_set_dma_mask(aPciDev, lBits);
    if (0 == lResult) { lResult = pci_set_consistent_dma_mask(aPciDev, lBits); }
#endif

    return lResult;
}

// aPciDev [---;RW-]
void at_pci_set_master(struct pci_dev *aPciDev)
{
    at_assert(NULL != aPciDev);
    pci_set_master(aPciDev);
}

// aPciDev [---;RW-]
// aSgList [---;RW-]
void at_pci_unmap_sg(struct pci_dev *aPciDev, void *aSgList, uint32_t aSgCount, bool aReadFromCard)
{
    DBG_PRINT("%s(%p, %p, %u, %u)\n", __FUNCTION__, aPciDev, aSgList, aSgCount, aReadFromCard);
    at_assert(NULL != aPciDev);
    at_assert(NULL != aSgList);
    at_assert(0 < aSgCount);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0)
    dma_unmap_sg(&aPciDev->dev, aSgList, aSgCount, aReadFromCard ? DMA_FROM_DEVICE : DMA_TO_DEVICE);
#else
    pci_unmap_sg(aPciDev, aSgList, aSgCount, aReadFromCard ? PCI_DMA_FROMDEVICE : PCI_DMA_TODEVICE);
#endif
}

// aPciDev [---;RW-]
int at_pci_write_config_byte(struct pci_dev *aPciDev, int aWhere, uint8_t aValue)
{
    at_assert(NULL != aPciDev);
    return pci_write_config_byte(aPciDev, aWhere, aValue);
}

// aPciDev [---;RW-]
int at_pci_write_config_dword(struct pci_dev *aPciDev, int aWhere, uint32_t aValue)
{
    at_assert(NULL != aPciDev);
    return pci_write_config_dword(aPciDev, aWhere, aValue);
}

// aEntries [---;R--]
unsigned int at_pci_msix_entry_get_entry(struct msix_entry *aEntries, unsigned int aIndex)
{
    at_assert(NULL != aEntries);
    return aEntries[aIndex].entry;
}

// aEntries [---;R--]
unsigned int at_pci_msix_entry_get_vector(struct msix_entry *aEntries, unsigned int aIndex)
{
    at_assert(NULL != aEntries);
    return aEntries[aIndex].vector;
}

// aEntries [---;-W-]
void at_pci_msix_entry_set_entry(struct msix_entry *aEntries,
                                 unsigned int aIndex,
                                 unsigned int aEntry)
{
    at_assert(NULL != aEntries);
    aEntries[aIndex].entry = aEntry;
}

// aFrom [---;R--]
struct pci_dev *
at_pci_retrieve_device(unsigned int aVendor, unsigned int aDevice, struct pci_dev *aFrom)
{
    return pci_get_device(aVendor, aDevice, aFrom);
}
