
/**
 * @file   linux/at/at_main.c
 *
 * @author Alazar Technologies Inc
 *
 * @copyright Copyright (c) 2020 Alazar Technologies Inc. All Rights
 * Reserved.  Unpublished - rights reserved under the Copyright laws
 * of the United States And Canada.
 * This product contains confidential information and trade secrets
 * of Alazar Technologies Inc. Use, disclosure, or reproduction is
 * prohibited without the prior express written permission of Alazar
 * Technologies Inc
 */

// This file is included into the dkms package.

// Includes
/////////////////////////////////////////////////////////////////////////////

// ===== Linux kernel =======================================================
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/version.h>

// ===== linux/at ===========================================================
#include <at_assert.h>
#include <at_linux_proc_fs.h>

#include <at_main.h>

// ===== linux/ats.../src ===================================================
#include <version.h>

// Public variables
/////////////////////////////////////////////////////////////////////////////

// QUESTION Cannot use GPL licence and use binary blob ?

MODULE_AUTHOR("Alazar Technologies Inc");
MODULE_DESCRIPTION("Alazar Technologies Driver");
MODULE_LICENSE("GPL");
MODULE_VERSION(VERSION_STRING);

// Static function declarations
/////////////////////////////////////////////////////////////////////////////

static int ProcFS_Open(struct inode *aInode, struct file *aFile);

// Static variables
/////////////////////////////////////////////////////////////////////////////

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
static struct proc_ops s_proc_ops = {
    .proc_open = ProcFS_Open,
    .proc_read = seq_read,
    .proc_write = at_procfs_write,
    .proc_lseek = seq_lseek,
    .proc_release = single_release,
};
#else
static struct file_operations sProcFS_FileOperations = {
    .open = ProcFS_Open,
    .read = seq_read,
    .write = at_procfs_write,
    .llseek = seq_lseek,
    .release = single_release,
};
#endif
// Functions
/////////////////////////////////////////////////////////////////////////////

void *at_compat_ptr(uint32_t aPtr)
{
#if defined(CONFIG_COMPAT_ENABLED) && defined(CONFIG_COMPAT)

    return compat_ptr(aPtr);

#else

    at_assert(false);

    return NULL;

#endif
}

// aThis    [---;-W-]
// aContext [--O;---]
void at_procfs_config(struct proc_dir_entry *aThis, void *aContext)
{
    at_assert(NULL != aThis);

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 10, 0)
    aThis->proc_fops = &sProcFS_FileOperations;
    aThis->data = aContext;
#endif
}

// aName    [---;R--]
// aParent  [---;RW-]
// aContext [-KO;---]
struct proc_dir_entry *at_procfs_create_data(const char *aName,
                                             umode_t aMode,
                                             struct proc_dir_entry *aParent,
                                             void *aContext)
{
    at_assert(NULL != aName);
    at_assert(NULL != aParent);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
    return proc_create_data(aName, aMode, aParent, &s_proc_ops, aContext);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3, 10, 0)
    return proc_create_data(aName, aMode, aParent, &sProcFS_FileOperations, aContext);
#else
    return create_proc_entry(aName, aMode, aParent);
#endif
}

// aOut    [---;-W-]
// aFormat [---;R--]
void at_sprintf(char *aOut, const char *aFormat, ...)
{
    va_list lArgs;

    at_assert(NULL != aOut);
    at_assert(NULL != aFormat);

    va_start(lArgs, aFormat);

    vsprintf(aOut, aFormat, lArgs);

    va_end(lArgs);
}

// Static functions
/////////////////////////////////////////////////////////////////////////////

// aInode [---;RW-]
// aFile  [---;RW-]
int ProcFS_Open(struct inode *aInode, struct file *aFile)
{
    at_assert(NULL != aInode);
    at_assert(NULL != aFile);

    return single_open(aFile, at_procfs_show, at_inode_get_data(aInode));
}
