
/**
 * @file   linux/at/at_main_A.c
 *
 * @author Alazar Technologies Inc
 *
 * @copyright Copyright (c) 2020 Alazar Technologies Inc. All Rights
 * Reserved.  Unpublished - rights reserved under the Copyright laws
 * of the United States And Canada.
 * This product contains confidential information and trade secrets
 * of Alazar Technologies Inc. Use, disclosure, or reproduction is
 * prohibited without the prior express written permission of Alazar
 * Technologies Inc
 */

// This file is included into the dkms package.

// Includes
/////////////////////////////////////////////////////////////////////////////

// ===== Linux kernel =======================================================
#include <linux/fs.h>
#include <linux/mod_devicetable.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/version.h>

// ===== linux/at ===========================================================
#include <at_assert.h>
#include <at_main.h>

#include <at_main_A.h>

// ===== src/ats... =========================================================
#include <DeviceId.h>

// Macros
/////////////////////////////////////////////////////////////////////////////

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 8, 0)
#define DEV_EXIT __devexit
#define DEV_INIT __devinit
#else
#define DEV_EXIT
#define DEV_INIT
#endif

// Static constants
/////////////////////////////////////////////////////////////////////////////

static struct pci_device_id sPciDeviceTable[] = {
    {VENDOR_ID, DEVICE_ID, PCI_ANY_ID, PCI_ANY_ID},
    {0},
};

// Public variables
/////////////////////////////////////////////////////////////////////////////

MODULE_DEVICE_TABLE(pci, sPciDeviceTable);

// Static function declarations
/////////////////////////////////////////////////////////////////////////////

static void __exit ModuleExit(void);
static int __init ModuleInit(void);

static int DEV_INIT Probe(struct pci_dev *aPciDev, const struct pci_device_id *aPciEntry);
static void DEV_EXIT Remove(struct pci_dev *aPciDev);

#ifdef CONFIG_PM
static int Resume(struct pci_dev *aPciDev);
static int Suspend(struct pci_dev *aPciDev, pm_message_t aState);
#endif

// ===== Old kernel support =================================================

#if defined(CONFIG_COMPAT_ENABLED) && defined(CONFIG_COMPAT)

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 35)
static int ConvertCompatIoCtl(struct inode *aINnode,
                              struct file *aFile,
                              unsigned int aCmd,
                              unsigned long aArgs);
#endif

#endif

// Module definition
/////////////////////////////////////////////////////////////////////////////

module_exit(ModuleExit);
module_init(ModuleInit);

// Static variables
/////////////////////////////////////////////////////////////////////////////

static struct file_operations sFileOperations;

static int sMajor;

static struct pci_driver sPciDriver = {
    .id_table = sPciDeviceTable,
    .probe = Probe,

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 8, 0)
    .remove = __devexit_p(Remove),
#else
    .remove = Remove,
#endif

#ifdef CONFIG_PM
    .resume = Resume,
    .suspend = Suspend,
#endif
};

// Static functions
/////////////////////////////////////////////////////////////////////////////

void __exit ModuleExit()
{
    // printk(KERN_DEBUG "ModuleExit()\n");

    pci_unregister_driver(&sPciDriver);

    unregister_chrdev(sMajor, DRIVER_NAME);

    sMajor = 0;
}

int __init ModuleInit()
{
    int lResult;

    // printk(KERN_DEBUG "ModuleInit()\n");

    sPciDriver.name = DRIVER_NAME;

    lResult = pci_register_driver(&sPciDriver);
    switch (lResult)
    {
        case -ENODEV:
            printk("pci_register_driver(  ) failed - ENODEV\n");
            pci_unregister_driver(&sPciDriver);
            break;
    }

    sFileOperations.owner = THIS_MODULE;

    sFileOperations.open = at_dispatch_open;
    sFileOperations.release = at_dispatch_release;
    sFileOperations.unlocked_ioctl = at_dispatch_ioctl;

#if (defined(CONFIG_COMPAT_ENABLED) && defined(CONFIG_COMPAT))
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 35)
    sFileOperations.compat_ioctl = ConvertCompatIoCtl;
#else
    sFileOperations.compat_ioctl = at_dispatch_compat_ioctl;
#endif
#endif

    sMajor = register_chrdev(0, DRIVER_NAME, &sFileOperations);

    return lResult;
}

// aPciDev   [-K-;RW-]
// aPciEntry [---;---]
int DEV_INIT Probe(struct pci_dev *aPciDev, const struct pci_device_id *aPciEntry)
{
    void *lDevExt;
    int lResult;

    // printk(KERN_DEBUG "Probe( , , )\n");

    at_assert(NULL != aPciDev);
    at_assert(NULL != aPciEntry);

    lDevExt = at_device_create(aPciDev);
    if (NULL == lDevExt)
    {
        printk(KERN_ERR "at_device_create(  ) failed\n");
        lResult = (-ENODEV);
    }
    else
    {
        pci_set_drvdata(aPciDev, lDevExt);

        lResult = 0;
    }

    return lResult;
}

// aPciDev [---;RW-]
void DEV_EXIT Remove(struct pci_dev *aPciDev)
{
    void *lDevExt;

    //printk(KERN_DEBUG "Remove(  )\n");

    at_assert(NULL != aPciDev);

    lDevExt = pci_get_drvdata(aPciDev);
    at_assert(NULL != lDevExt);

    pci_set_drvdata(aPciDev, NULL);

    at_device_release(lDevExt);
}

#ifdef CONFIG_PM

// aPciDev [---;RW-]
int Resume(struct pci_dev *aPciDev)
{
    // printk(KERN_DEBUG "Resume(  )\n");

    at_assert(NULL != aPciDev);

    pci_set_power_state(aPciDev, PCI_D0);
    pci_restore_state(aPciDev);

    return 0;
}

// aPciDev [---;RW-]
int Suspend(struct pci_dev *aPciDev, pm_message_t aState)
{
    // printk(KERN_DEBUG "Suspend( ,  )\n");

    at_assert(NULL != aPciDev);

    pci_save_state(aPciDev);
    pci_set_power_state(aPciDev, pci_choose_state(aPciDev, aState));

    return 0;
}

#endif

// ===== Old kernel support =================================================

#if defined(CONFIG_COMPAT_ENABLED) && defined(CONFIG_COMPAT)

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 35)

// aINode [---;---]
// aFile  [---;---]
int ConvertCompatIoCtl(struct inode *aINnode,
                       struct file *aFile,
                       unsigned int aCmd,
                       unsigned long aArgs)
{
    at_assert(NULL != aINode);
    at_assert(NULL != aFile);

    at_dispatch_compat_ioctl(aFile, aCmd, aArgs);
}

#endif

#endif
