/**
 * @file
 *
 * @author Alazar Technologies Inc
 *
 * @copyright Copyright (c) 2019 Alazar Technologies Inc. All Rights
 * Reserved.  Unpublished - rights reserved under the Copyright laws
 * of the United States And Canada.
 * This product contains confidential information and trade secrets
 * of Alazar Technologies Inc. Use, disclosure, or reproduction is
 * prohibited without the prior express written permission of Alazar
 * Technologies Inc
 */

#ifndef ATS_ERROR_H
#define ATS_ERROR_H

#include "platform.h"

typedef enum
{
    rc_success = 0,
    rc_failed = -100,
    rc_timeout = -101,
    rc_invalid = -102,
    rc_invalid_index = -103,
    rc_device_configuration_error = -104,
    rc_unsupported_parameter = -105,
    rc_invalid_data = -106,
    rc_null_param = -107,
    rc_pll_not_locked = -108,
    rc_spi_invalid_signal = -109,      ///< Wrong member of `spi_signal_t` used
    rc_spi_invalid_line = -110,        ///< Invalid `spi_line_t` object used
    rc_spi_timeout = -111,             ///< The operation timed out
    rc_spi_insufficient_length = -112, ///< Length parameter is too small
    rc_spi_invalid_options = -113,     ///< Invalid options used
    rc_spi_null_param = -114,          ///< A parameter passed is `null`
    rc_spi_reg_access_failed = -115,   ///< Register access failed
    rc_spi_invalid_config = -116,      ///< Invalid `spi_config_t` object is used.
    rc_crc_invalid = -117,             ///< The CRC of a board database is invalid.
    rc_alloc_failed = -118,            ///< Memory allocation failed
    rc_unsupported_function = -119,
    rc_invalid_size = -120,
    rc_invalid_input_range = -121,
    rc_invalid_cache = -122,
    rc_idle = -123,
    rc_busy = -124,
    rc_done = -125,
    rc_reference_clock_pll_not_locked = -126,
    rc_adc_jesd_pll_not_locked = -127,
    rc_adc_jesd_link_not_locked = -128,
    rc_fpga_jesd_link_not_locked = -129,
} rc_t;

static inline const char *rc_name(rc_t rc)
{
    switch (rc)
    {
        case rc_success:
            return "success";
        case rc_failed:
            return "failed";
        case rc_timeout:
            return "timeout";
        case rc_invalid:
            return "invalid";
        case rc_invalid_index:
            return "invalid index";
        case rc_device_configuration_error:
            return "device_configuration_error";
        case rc_unsupported_parameter:
            return "unsupported_parameter";
        case rc_invalid_data:
            return "invalid data";
        case rc_null_param:
            return "null parameter";
        case rc_pll_not_locked:
            return "PLL not locked";
        case rc_spi_invalid_signal:
            return "Invalid SPI signal";
        case rc_spi_invalid_line:
            return "Invalid SPI line";
        case rc_spi_timeout:
            return "SPI timeout";
        case rc_spi_insufficient_length:
            return "SPI error, insufficient length";
        case rc_spi_invalid_options:
            return "Invalid SPI options";
        case rc_spi_null_param:
            return "Null SPI param";
        case rc_spi_reg_access_failed:
            return "SPI register access failed";
        case rc_spi_invalid_config:
            return "Invalid SPI configuration";
        case rc_crc_invalid:
            return "Invalid CRC";
        case rc_alloc_failed:
            return "Memory allocation failed";
        case rc_unsupported_function:
            return "Unsupported function";
        case rc_invalid_size:
            return "Invalid size";
        case rc_invalid_input_range:
            return "Invalid input range";
        case rc_invalid_cache:
            return "Invalid cache";
        case rc_idle:
            return "Idle";
        case rc_busy:
            return "Busy";
        case rc_done:
            return "Done";
        case rc_reference_clock_pll_not_locked:
            return "Reference clock PLL not locked";
        case rc_adc_jesd_pll_not_locked:
            return "ADC JESD PLL not locked";
        case rc_adc_jesd_link_not_locked:
            return "ADC JESD link not locked";
        case rc_fpga_jesd_link_not_locked:
            return "FPGA JESD link not locked";
        default:
            return "unknown";
    }
}

static inline rc_t ats_internal_error(const char *file, const char *func, int line, rc_t rc)
{
    if (rc != rc_success)
        ats_log_error("Error: %s:%s:%d: %s (%d)\n", file, func, line, rc_name(rc), rc);

    return rc;
}

#define ats_error(__r) ats_internal_error(__FILE__, __FUNCTION__, __LINE__, __r)

/// Executes the expression `__x`, and checks that it evaluates to `rc_success`.
/// If it doesn't, logs an error using `ats_log_error` and calls `return` with
/// the value of `__x`.
#define ATS_CHECK(__x)                                                                             \
    do                                                                                             \
    {                                                                                              \
        rc_t __r = (__x);                                                                          \
        if (__r != rc_success)                                                                     \
        {                                                                                          \
            ats_log_error("[%s] %s failed : %d\n", __FUNCTION__, #__x, __r);                       \
            return __r;                                                                            \
        }                                                                                          \
    } while (0)

/// Checks if `__x` is `NULL`. If it is, logs an error using `ats_log_error` and
/// returns `rc_null_param`.
#define ATS_CHECK_NULL(__x)                                                                        \
    do                                                                                             \
    {                                                                                              \
        if (__x == NULL)                                                                           \
        {                                                                                          \
            ats_log_error("[%s] %s is NULL\n", __FUNCTION__, #__x);                                \
            return rc_null_param;                                                                  \
        }                                                                                          \
    } while (0)

/// Used with ATS_ASSERT and ATS_CHECK_EXP
#define NO_RETURN_VALUE

/// Checks if the expression `__x` is evaluated to `true`.
/// If it doesn't, logs an error using `ats_log_error` and calls `return` with
/// the value of `__r`.
/// To use this macro in a function that returns `void`, please use `NO_RETURN_VALUE`.
#define ATS_CHECK_EXP(__x, __r)                                                                    \
    do                                                                                             \
    {                                                                                              \
        if (!(__x))                                                                                \
        {                                                                                          \
            ats_log_error("[%s] %s is false", __FUNCTION__, #__x);                                 \
            return __r;                                                                            \
        }                                                                                          \
    } while (0)

/// Equivalent of the C `assert()` function. This only generates code in
/// debug mode. In case of failure, instead of terminating the application
/// like assert() does, this macro returns the value passed in `r`.
/// Checks if the expression `__x` is evaluated to `true`.
/// If it doesn't, logs an error using `ats_log_error` and calls `return` with
/// the value of `__r`.
/// To use this macro in a function that returns `void`, please use `NO_RETURN_VALUE`.
#ifdef NDEBUG
#define ATS_ASSERT(__x, __r)
#else
#define ATS_ASSERT(__x, __r) ATS_CHECK_EXP(__x, __r)
#endif

/// Log the execution time of the expression `__x`.
#define ATS_EXECUTION_TIME(__x)                                                                    \
    do                                                                                             \
    {                                                                                              \
        int64_t __s = ats_get_utime();                                                             \
        __x;                                                                                       \
        ats_log_trace("[%s] execution time %lldus\n", #__x, ats_get_utime() - __s);                \
    } while (0)

#if defined(_DRIVER_) && defined(_WIN32)

static inline NTSTATUS rc_to_nt_status(rc_t rc)
{
    switch (rc)
    {
        case rc_success:
            return STATUS_SUCCESS;

        case rc_failed:
        case rc_alloc_failed:
        case rc_device_configuration_error:
        case rc_pll_not_locked:
            return STATUS_UNSUCCESSFUL;

        case rc_invalid:
        case rc_invalid_size:
        case rc_unsupported_function:
            return STATUS_INVALID_DEVICE_REQUEST;

        case rc_null_param:
        case rc_invalid_index:
        case rc_invalid_data:
        case rc_invalid_input_range:
        case rc_unsupported_parameter:
            return STATUS_INVALID_PARAMETER;
    }

    return STATUS_SUCCESS;
}

#endif

#endif // ATS_ERROR_H
