//=============================================================================
//
// Alazar Technologies Inc
//
// Copyright (c) 2019 Alazar Technologies Inc. All Rights Reserved.
// Unpublished - rights reserved under the Copyright laws of the
// United States And Canada.
//
// This product contains confidential information and trade secrets
// of Alazar Technologies Inc. Use, disclosure, or reproduction is
// prohibited without the prior express written permission of Alazar
// Technologies Inc
//
//=============================================================================

#include <linux/string.h>
#include <linux/version.h>

#include <crypto/hash.h>

#include "platform.h"

#include "ats_hash.h"

#include "utils.h"

/******************************************************************************
 *
 * Function   :  ats_hash_open
 *
 * Description:  Open hash handle
 *
 ******************************************************************************/
ats_hash_t ats_hash_open(ats_hash_algo_t algo)
{
    struct crypto_shash *palgo;
    char *algo_id;

    switch (algo)
    {
        case ATS_HASH_MD5:
            algo_id = "md5";
            break;

        case ATS_HASH_SHA1:
            algo_id = "sha1";
            break;

        case ATS_HASH_SHA256:
            algo_id = "sha256";
            break;

        default:
            return NULL;
    }

    palgo = crypto_alloc_shash(algo_id, CRYPTO_ALG_TYPE_SHASH, 0);
    return (ats_hash_t)palgo;
}

/******************************************************************************
 *
 * Function   :  ats_hash_close
 *
 * Description:  Close hash handle
 *
 ******************************************************************************/
ats_hash_digest_t ats_hash_close(ats_hash_t handle)
{
    struct crypto_shash *palgo = (struct crypto_shash *)handle;
    ats_hash_digest_t digest = {.status = -1, .algo = ATS_HASH_NONE};

    crypto_free_shash(palgo);
    return digest;
}

/******************************************************************************
 *
 * Function   :  ats_hash_add
 *
 * Description:  Add data to hash computation
 *
 ******************************************************************************/
int32_t ats_hash_add(ats_hash_t handle, uint8_t *pdata, uint32_t size)
{
    struct crypto_shash *palgo = (struct crypto_shash *)handle;
    ats_hash_digest_t digest = {.status = -1, .algo = ATS_HASH_NONE};
    struct shash_desc shash =
    {
        .tfm = palgo,

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 2, 0)
        .flags = 0,
#endif
    };

    switch (digest.algo)
    {
        case ATS_HASH_MD5:
            return crypto_shash_digest(&shash, pdata, size, digest.u.md5);

        case ATS_HASH_SHA1:
            return crypto_shash_digest(&shash, pdata, size, digest.u.sha1);

        case ATS_HASH_SHA256:
            return crypto_shash_digest(&shash, pdata, size, digest.u.sha256);

        default:
            break;
    }

    return -1;
}

/******************************************************************************
 *
 * Function   :  ats_hash_compare
 *
 * Description:  Compare 2 digests
 *
 ******************************************************************************/
int ats_hash_compare(ats_hash_digest_t left, ats_hash_digest_t right)
{
    if ((left.status < 0) || (right.status < 0) || (left.algo != right.algo)) return -1;

    switch (left.algo)
    {
        case ATS_HASH_MD5:
            return ((memcmp(left.u.md5, right.u.md5, sizeof(left.u.md5)) == 0) ? 0 : -1);

        case ATS_HASH_SHA1:
            return ((memcmp(left.u.sha1, right.u.sha1, sizeof(left.u.sha1)) == 0) ? 0 : -1);

        case ATS_HASH_SHA256:
            return ((memcmp(left.u.sha256, right.u.sha256, sizeof(left.u.sha256)) == 0) ? 0 : -1);

        default:
            return -1;
    }

    return -1;
}

/******************************************************************************
 *
 * Function   :  ats_hash_log
 *
 * Description:  Log digest message
 *
 ******************************************************************************/
void ats_hash_log(ats_hash_digest_t digest, char *pmsg)
{
    char buffer[sizeof(digest.u) * 2 + 1];
    char *palgo;
    char *pbuf;
    uint8_t *pdig;
    int len = 0;
    int idx;

    switch (digest.algo)
    {
        case ATS_HASH_MD5:
            pdig = digest.u.md5;
            len = sizeof(digest.u.md5);
            palgo = "MD5";
            break;

        case ATS_HASH_SHA1:
            pdig = digest.u.sha1;
            len = sizeof(digest.u.sha1);
            palgo = "SHA1";
            break;

        case ATS_HASH_SHA256:
            pdig = digest.u.sha256;
            len = sizeof(digest.u.sha256);
            palgo = "SHA256";
            break;

        default:
            return;
    }

    for (pbuf = buffer, idx = 0; idx < len; idx++)
        pbuf += ats_snprintf(pbuf, sizeof(buffer), "%02x", pdig[idx]);

    if (pmsg == NULL)
        ats_log_info("%s=%s\n", palgo, buffer);
    else
        ats_log_info("%s %s=%s\n", pmsg, palgo, buffer);
}
