/**
 * @file
 *
 * @author Alazar Technologies Inc
 *
 * @copyright Copyright (c) 2019 Alazar Technologies Inc. All Rights
 * Reserved.  Unpublished - rights reserved under the Copyright laws
 * of the United States And Canada.
 * This product contains confidential information and trade secrets
 * of Alazar Technologies Inc. Use, disclosure, or reproduction is
 * prohibited without the prior express written permission of Alazar
 * Technologies Inc
 */

#ifndef ATS_HASH_H
#define ATS_HASH_H

#include "ats_types.h"

#include "platform.h"

///< Crypto hash algorythm
typedef enum
{
    ATS_HASH_NONE,
    ATS_HASH_MD5,
    ATS_HASH_SHA1,
    ATS_HASH_SHA256,
    ATS_HASH_MAX
} ats_hash_algo_t;

typedef void *ats_hash_t; ///< Handle hash type

typedef struct
{
    /// Status of the hashing operation. Other fields are valid if and only if
    /// this is zero.
    int status;

    /// Hashing algorithm used. This defines which field of `u` is valid.
    ats_hash_algo_t algo;

    /// Message digest
    union {
        uint8_t md5[16];
        uint8_t sha1[20];
        uint8_t sha256[32];
    } u;
} ats_hash_digest_t;

/**
 * @brief Open hash computing descriptor
 *
 * @param algo      The hash algorythm type
 *
 * @param pres      Pointer to the hash result buffer
 *
 * @param size      Size of the hash result buffer
 *
 */
ats_hash_t ats_hash_open(ats_hash_algo_t algo);

/**
 * @brief Close hash computing descriptor
 *
 * @param handle    The hash computing handle returned by ats_hash_open()
 *
 */
ats_hash_digest_t ats_hash_close(ats_hash_t handle);

/**
 * @brief Add data to hash computation
 *
 * @param handle    The hash computing handle returned by ats_hash_open()
 *
 * @param pdata     Pointer to a buffer of data to add to hash computation
 *
 * @param size      Size of the data to add to hash computation
 *
 */
int ats_hash_add(ats_hash_t handle, uint8_t *pbuf, uint32_t size);

/**
 * @brief Compare digest structures
 *
 * @param left      Left digest structure to compare
 *
 * @param right     Right digest structure to compare
 *
 */
int ats_hash_compare(ats_hash_digest_t left, ats_hash_digest_t right);

/**
 * @brief Compare digest structures
 *
 * @param digest    Digest to log
 *
 * @param pmsg      Prefix message string
 *
 */
void ats_hash_log(ats_hash_digest_t digest, char *pmsg);

#endif // ATS_HASH_H
