
#include "ats_pci.h"

static rc_t ats_pci_config_read(void *buffer, size_t offset, size_t length, void *context)
{
    ATS_ASSERT(buffer != NULL, rc_null_param);
    ATS_ASSERT(context != NULL, rc_null_param);

    driver_os_handle_t *driver = (driver_os_handle_t *)context;

    for (; length > 0; length--, offset++, buffer++)
    {
        int rc = at_pci_read_config_byte(driver->device, offset, (uint8_t *)buffer);

        if (rc)
        {
            ats_log_error("[%s] at_pci_read_config_byte failed: %d\n", __FUNCTION__, rc);
            return rc_failed;
        }
    }

    return rc_success;
}

static rc_t ats_pci_config_write(void *buffer, size_t offset, size_t length, void *context)
{
    ATS_ASSERT(buffer != NULL, rc_null_param);
    ATS_ASSERT(context != NULL, rc_null_param);

    driver_os_handle_t *driver = (driver_os_handle_t *)context;

    for (; length > 0; length--, offset++, buffer++)
    {
        int rc = at_pci_write_config_byte(driver->device, offset, *((uint8_t *)buffer));

        if (rc)
        {
            ats_log_error("[%s] at_pci_write_config_byte failed: %d\n", __FUNCTION__, rc);
            return rc_failed;
        }
    }

    return rc_success;
}

static ats_pci_irq_t *ats_pci_get_irq(ats_pci_t *ats_pci, int irq)
{
    if ((irq < 0) || (irq >= (int)(ats_pci->irq_count))) return NULL;

    return &ats_pci->irqs[irq];
}

static ats_pci_irq_t *ats_pci_get_irq_from_vector(ats_pci_t *ats_pci, int vector)
{
    for (int irq = 0; irq < ats_pci->irq_count; irq++)
    {
        ats_pci_irq_t *ats_irq = &ats_pci->irqs[irq];

        if (ats_irq->vector == vector) return ats_irq;
    }

    return NULL;
}

static void ats_pci_tasklet(unsigned long context)
{
    ats_pci_irq_t *irq = (ats_pci_irq_t *)context;

    if ((irq == NULL) || (irq->handler == NULL)) return;

    irq->handler(irq);
}

static irqreturn_t ats_pci_interrupt(int vector, void *context)
{
    ats_pci_t *ats_pci = (ats_pci_t *)context;

    if (ats_pci == NULL) return IRQ_HANDLED;

    ats_pci_irq_t *irq = ats_pci_get_irq_from_vector(ats_pci, vector);

    if ((irq != NULL) && (irq->handler != NULL))
        at_tasklet_schedule_gated(&irq->tasklet, &ats_pci->tasklet_gate);

    return IRQ_HANDLED;
}

static rc_t ats_pci_open_internal(ats_pci_t *ats_pci, driver_os_handle_t *driver, bool with_irqs)
{
    // Setup prior to scanning hardware resources
    ats_pci_init(ats_pci, ats_pci_config_read, ats_pci_config_write, driver);

    for (int index = 0; index < ATS_PCI_BAR_MAX; index++)
    {
        ats_pci_bar_t *bar = &ats_pci->bars[index];

        if ((bar->length = at_pci_resource_len(driver->device, index)) == 0) continue;

        if (at_pci_resource_io(driver->device, index))
        {
            ats_log_error("[%s] PCI I/O resources are not supported\n", __FUNCTION__);
            continue;
        }

        bar->type = ATS_PCI_BAR_TYPE_MEMORY;

        if ((bar->virtual_address = at_pci_ioremap(driver->device, index)) == NULL)
        {
            ats_log_error("[%s] Unable to map memory region for bar %d\n", __FUNCTION__, index);
            return rc_failed;
        }
    }

    if ((!with_irqs) || (ats_pci->irq_count == 0)) return rc_success;

    if ((ats_pci->irqs = (ats_pci_irq_t *)ats_alloc(sizeof(ats_pci_irq_t) * ats_pci->irq_count)) ==
        NULL)
        return rc_alloc_failed;

    if (ats_pci->msix_enabled)
    {
        for (int index = 0; index < ats_pci->irq_count; index++)
            at_pci_msix_entry_set_entry(ats_pci->msix_entries, index, index);

        if (!at_pci_alloc_irq_vectors_MSIX(driver->device, ats_pci->irq_count,
                                           ats_pci->msix_entries))
            return rc_failed;
    }
    else
    {
        if (!at_pci_alloc_irq_vectors_MSI(driver->device, ats_pci->irq_count)) return rc_failed;

        at_pci_enable_msi(driver->device);
    }

    for (int index = 0; index < ats_pci->irq_count; index++)
    {
        ats_pci_irq_t *irq = &ats_pci->irqs[index];

        irq->user_id = -1;
        irq->vector = at_pci_irq_vector(driver->device, index);
        irq->handler = NULL;
        irq->on_enable_request = NULL;
        irq->context = NULL;
        at_tasklet_init(&irq->tasklet, ats_pci_tasklet, (unsigned long)irq);
    }

    return rc_success;
}

rc_t ats_pci_open(ats_pci_t *ats_pci, driver_os_handle_t *driver)
{
    return ats_pci_open_internal(ats_pci, driver, false);
}

rc_t ats_pci_open_with_irqs(ats_pci_t *ats_pci, driver_os_handle_t *driver)
{
    return ats_pci_open_internal(ats_pci, driver, true);
}

rc_t ats_pci_close(ats_pci_t *ats_pci)
{
    if (ats_pci->irqs != NULL)
    {
        if (ats_pci->msix_enabled)
            at_pci_disable_msix(ats_pci->driver.device);
        else
            at_pci_disable_msi(ats_pci->driver.device);

        for (int index = 0; index < ats_pci->irq_count; index++)
        {
            ats_pci_irq_t *irq = &ats_pci->irqs[index];

            at_tasklet_kill(&irq->tasklet);
            at_irq_free(irq->vector, irq);
        }

        ats_free(ats_pci->irqs);
        ats_pci->irq_count = 0;
        ats_pci->irqs = NULL;
    }

    for (int index = 0; index < ATS_PCI_BAR_MAX; index++)
    {
        ats_pci_bar_t *bar = &ats_pci->bars[index];

        if ((bar->type != ATS_PCI_BAR_TYPE_MEMORY) || (bar->length == 0) ||
            (bar->virtual_address == NULL))
            continue;

        at_pci_iounmap(bar->virtual_address);
        bar->virtual_address = NULL;
        bar->length = 0;
    }

    return rc_success;
}

rc_t ats_pci_register_irq(ats_pci_t *ats_pci,
                          int irq,
                          interrupt_handler_t handler,
                          on_interrupt_enable_request_t on_enable_request,
                          void *context)
{
    if (handler == NULL) return rc_invalid_data;

    if (ats_pci->irqs == NULL) return rc_failed;

    ats_pci_irq_t *ats_irq;

    if ((ats_irq = ats_pci_get_irq(ats_pci, irq)) == NULL) return rc_invalid_index;

    ats_irq->handler = handler;
    ats_irq->on_enable_request = on_enable_request;
    ats_irq->context = context;

    if (at_irq_request(ats_irq->vector, ats_pci_interrupt, false, ats_pci->driver.name, ats_pci))
        return rc_success;

    return rc_failed;
}

rc_t ats_pci_enable_irq(ats_pci_t *ats_pci, int irq)
{
    ats_pci_irq_t *ats_irq;

    if ((ats_irq = ats_pci_get_irq(ats_pci, irq)) == NULL) return rc_invalid_index;

    at_irq_enable(ats_irq->vector);
    return rc_success;
}

rc_t ats_pci_disable_irq(ats_pci_t *ats_pci, int irq)
{
    ats_pci_irq_t *ats_irq;

    if ((ats_irq = ats_pci_get_irq(ats_pci, irq)) == NULL) return rc_invalid_index;

    at_irq_disable(ats_irq->vector);
    return rc_success;
}
