///
/// @file
///
/// This module is a cross-platform interface for the base address registers (BARs) and the
/// configuration space of a PCI device. This module is centered around an ::ats_pci_t context
/// and exposes data access function like ::ats_pci_bar_write_u8 and ::ats_pci_bar_read_u32.
///
/// Since mapping in virtual memory of BARs is done differently depending on the operating system,
/// there is no `ats_pci_open` function. Instead, ::ats_pci_t contexts must be created by
/// platform-specific modules like `ats_pci_nt`.
///
/// @author Alazar Technologies Inc
///
/// @copyright Copyright (c) 2020 Alazar Technologies Inc. All Rights Reserved.
///

// This file is included into the dkms package.

#ifndef ATS_PCI_H
#define ATS_PCI_H

#include "ats_error.h"
#include "ats_types.h"

#ifdef _WIN32
#include "wdf.h"
#else
#include "at_gate.h"
#include "at_linux_interrupt.h"
#include "at_linux_pci.h"
#endif

#include "ats_os.h"

#include "atssys.h"

/// Represents the type of a BAR region
typedef enum
{
    ATS_PCI_BAR_TYPE_IO,     ///< I/O region. This is deprecated
    ATS_PCI_BAR_TYPE_MEMORY, ///< Memory region.

    /// @cond INTERNAL_DECLARATIONS
    /// Must be the last element of the enum
    ATS_PCI_BAR_TYPE_MAX,
    /// @endcond INTERNAL_DECLARATIONS
} ats_pci_bar_type_t;

/// A memory region mapped to a PCI device
typedef enum
{
    ATS_PCI_BAR_0, ///< PCI BAR 0
    ATS_PCI_BAR_1, ///< PCI BAR 1
    ATS_PCI_BAR_2, ///< PCI BAR 2
    ATS_PCI_BAR_3, ///< PCI BAR 3
    ATS_PCI_BAR_4, ///< PCI BAR 4
    ATS_PCI_BAR_5, ///< PCI BAR 5

    /// @cond INTERNAL_DECLARATIONS
    /// Must be the last element of the enum
    ATS_PCI_BAR_MAX
    /// @endcond INTERNAL_DECLARATIONS
} ats_pci_bar_id_t;

/// Indicates whether a PCI register field can be read from, written to, or both.
typedef enum ats_pci_regfield_mode
{
    ATS_PCI_REGFIELD_MODE_READWRITE = 0,
    ATS_PCI_REGFIELD_MODE_WRITEONLY = 0x00000001UL,
    ATS_PCI_REGFIELD_MODE_READONLY = 0x00000002UL,
} ats_pci_regfield_mode_t;

#pragma pack(push, 1)

typedef struct ats_pci_bar_regfield_s
{
    ats_pci_bar_id_t bar;
    uint32_t address;
    int8_t offset;
    int8_t length;
    ats_pci_regfield_mode_t mode;
} ATTR_PACKED ats_pci_bar_regfield_t;

typedef struct ats_pci_config_regfield_s
{
    uint32_t address;
    int8_t offset;
    int8_t length;
    ats_pci_regfield_mode_t mode;
} ATTR_PACKED ats_pci_config_regfield_t;

#pragma pack(pop)

/// Return `true` if `field` is not obviously invalid. Examples of obviously invalid fields are
/// zero-initialized structs and fields for which `offset + length > 32`
static inline bool ats_pci_bar_regfield_is_valid(ats_pci_bar_regfield_t field)
{
    return ((field.offset >= 0) && (field.length > 0) && (field.offset + field.length <= 32));
}

/// Return `true` if `field` is not obviously invalid. Examples of obviously invalid fields are
/// zero-initialized structs and fields for which `offset + length > 32`
static inline bool ats_pci_config_regfield_is_valid(ats_pci_config_regfield_t field)
{
    return ((field.offset >= 0) && (field.length > 0) && (field.offset + field.length <= 32));
}

/// Represents a BAR region mapped in virtual memory.
///
/// @note This structure is copyable
typedef struct
{
    ats_pci_bar_id_t id;     ///< Bar id
    ats_pci_bar_type_t type; ///< Type of memory region that this object represents
    int64_t length;          ///< Size of the region in bytes
    void *virtual_address;   ///< Virtual address at which the region is mapped
    bool is_64bits;          ///< Set to true for 64 bits Bars
} ats_pci_bar_t;

/// Reads a data buffer from a PCI device's configuration space.
///
/// This function is implemented by modules that initialize ::ats_pci_t contexts (e.g.
/// ats_pci_nt). It is used exclusively in the implementation of the ats_pci module. Users of this
/// module should not call this function directly, and instead call functions like
/// ::ats_pci_t_config_read_u8
///
/// @param buffer        Pointer to the data buffer to read
/// @param offset_bytes  Offset in the PCI config space where data should be read
/// @param length_bytes  Number of bytes to write
/// @param user_data     ::ats_pci_t::user_data gets passed to this parameter when this function
///                      is called
typedef rc_t (*ats_pci_config_read_f)(void * /* buffer */,
                                      size_t /* offset_bytes */,
                                      size_t /* length_bytes */,
                                      void * /* user_data */);

/// Writes a data buffer to a PCI device's configuration space.
///
/// This function is implemented by modules that initialize ::ats_pci_t contexts (e.g.
/// ats_pci_nt). It is used exclusively in the implementation of the ats_pci module. Users of this
/// module should not call this function directly, and instead call functions like
/// ::ats_pci_config_write_u8
///
/// @param buffer        Pointer to the data buffer to write
/// @param offset_bytes  Offset in the PCI config space where data should be written
/// @param length_bytes  Number of bytes to write
/// @param user_data     ::ats_pci_t::user_data gets passed to this parameter when this function
///                      is called
typedef rc_t (*ats_pci_config_write_f)(void * /* buffer */,
                                       size_t /* offset_bytes */,
                                       size_t /* length_bytes */,
                                       void * /* user_data */);

struct ats_pci_s;
struct ats_pci_irq_s;

typedef void (*interrupt_handler_t)(struct ats_pci_irq_s *);
typedef void (*on_interrupt_enable_request_t)(struct ats_pci_irq_s *, bool);

/// Represents an MSIX PCI interrupt entry in the MSIX table.
typedef struct ats_pci_msix_s
{
    uint64_t address;
    uint32_t data;
    uint32_t control;
} ATTR_PACKED ats_pci_msix_t;

/// Represents an MSI or MSIX PCI interrupt request.
///
/// @note This structure is copyable
typedef struct ats_pci_irq_s
{
    int user_id;                                     ///< Interrupt user Id number
    int vector;                                      ///< Interrupt vector Id number
    interrupt_handler_t handler;                     ///< Interrupt handler
    on_interrupt_enable_request_t on_enable_request; ///< Callback used to enable/disable interrupt
    void *context; ///< Context used by handler and on_enable_request

    /// Platform dependent fields
#ifdef _WIN32
    WDFINTERRUPT interrupt;
    unsigned long resource;
#else
    struct tasklet_struct tasklet;
#endif
} ats_pci_irq_t;

typedef struct ats_pci_common_config_s
{
    uint16_t vendor_id;
    uint16_t device_id;
    uint16_t command;
    uint16_t status;
    uint8_t revision_id;
    uint8_t prog_if;
    uint8_t sub_class;
    uint8_t base_class;
    uint8_t cache_line_size;
    uint8_t latency_timer;
    uint8_t header_type;
    uint8_t BIST;
    uint32_t raw_base_addresses[6];
    uint32_t card_bus_CIS_pointer;
    uint16_t subsystem_vendor_id;
    uint16_t subsystem_id;
    uint32_t ROM_base_address;
    uint8_t capabilities_pointer;
    uint8_t reserved1[3];
    uint32_t reserved2;
    uint8_t interrupt_line;
    uint8_t interrupt_pin;
    uint8_t minimum_grant;
    uint8_t maximum_latency;
    uint8_t device_specific[192];
} ATTR_PACKED ats_pci_common_config_t;

typedef struct ats_pci_capabilities_header_s
{
    uint8_t id;
    uint8_t next;
} ATTR_PACKED ats_pci_capability_header_t;

typedef struct ats_pci_express_capabilities_s
{
    ats_pci_capability_header_t header;
    uint16_t capabilities_register;
    uint32_t device_capabilities;
    uint16_t device_control;
    uint16_t device_status;
    uint32_t link_capabilities;
    uint16_t link_control;

    struct pci_express_link_status_s
    {
        uint16_t link_speed : 4;
        uint16_t link_width : 6;
        uint16_t link_tranining_error : 1;
        uint16_t link_tranining : 1;
        uint16_t slot_clock_configuration : 1;
        uint16_t data_link_layer_active : 1;
        uint16_t reserved : 2;
    } link_status;

    uint32_t slot_capabilities;
    uint16_t slot_control;
    uint16_t slot_status;
    uint16_t root_control;
    uint16_t reserved;
    uint32_t root_status;
} ATTR_PACKED ats_pci_express_capabilities_t;

/// Main context of the ats_pci module.
///
/// @note This structure is copyable
typedef struct ats_pci_s
{
    /// Reads data from the PCI config space
    ats_pci_config_read_f config_read;

    /// Writes data to the PCI config space
    ats_pci_config_write_f config_write;

    /// Passed as the last argument to ::config_read and ::config_write
    void *user_data;

    /// This array contains data relative to the BAR regions.
    ats_pci_bar_t bars[ATS_PCI_BAR_MAX];

    /// Set to true if device MSIX are enabled
    bool msix_enabled;

    /// Number of active message interrupts on this device
    size_t irq_count;

    /// Pointer to an allocated interrupt request array.
    ats_pci_irq_t *irqs;

    /// Platform driver context
    driver_os_handle_t driver;

    /// PCI common config cache
    ats_pci_common_config_t config;

    /// PCI express capabilities cache
    ats_pci_express_capabilities_t capabilities;

    /// Platform dependent fields
#ifdef _WIN32
#else
    struct at_gate_rt tasklet_gate;
    struct msix_entry *msix_entries;
#endif
} ats_pci_t;

/// Discovers memory resources for a PCI device, maps them to virtual memory and initializes
/// `ats_pci`.
///
/// @param ats_pci     Pass a pointer to an uninitialized `ats_pci_t` struct. This function will
///                    initialize its contents.
/// @param driver      Operating system driver context
rc_t ats_pci_open(ats_pci_t *ats_pci, driver_os_handle_t *driver);

/// Discovers memory and interrupts resources for a PCI device, maps them to virtual memory and
/// initializes `ats_pci`.
///
/// @param ats_pci     Pass a pointer to an uninitialized `ats_pci_t` struct. This function will
///                    initialize its contents.
/// @param driver      Operating system driver context
rc_t ats_pci_open_with_irqs(ats_pci_t *ats_pci, driver_os_handle_t *driver);

/// Unmaps resources for a PCI device
///
/// @param ats_pci  Pointer to the device for which to unmap memory
rc_t ats_pci_close(ats_pci_t *ats_pci);

/// Writes a buffer of `uint8_t` values to a BAR
///
/// @param ats_pci   PCI device to write to
/// @param space     BAR to write to
/// @param offset    Start offset in `uint8_t` in the BAR
/// @param data      Pointer to the buffer to write
/// @param count     Number of `uint8_t` values to write
rc_t ats_pci_bar_write_u8(
    ats_pci_t *ats_pci, ats_pci_bar_id_t id, int offset, uint8_t *data, size_t count);

/// Reads a buffer of `uint8_t` values from a BAR
///
/// @param ats_pci   PCI device to read from
/// @param space     BAR to read from
/// @param offset    Start offset in `uint8_t` in the BAR
/// @param data      Pointer to the buffer to read
/// @param count     Number of `uint8_t` values to read
rc_t ats_pci_bar_read_u8(
    ats_pci_t *ats_pci, ats_pci_bar_id_t id, int offset, uint8_t *data, size_t count);

/// Writes a buffer of `uint16_t` to a BAR
///
/// @param ats_pci   PCI device to write to
/// @param space     BAR to write to
/// @param offset    Start offset in `uint16_t` in the BAR
/// @param data      Pointer to the buffer to write
/// @param count     Number of `uint16_t` values to write
rc_t ats_pci_bar_write_u16(
    ats_pci_t *ats_pci, ats_pci_bar_id_t id, int offset, uint16_t *data, size_t count);

/// Reads a buffer of uint16_t from a BAR
///
/// @param ats_pci   PCI device to write to
/// @param space     BAR to read from
/// @param offset    Start offset in `uint16_t` in the BAR
/// @param data      Pointer to the buffer to read
/// @param count     Number of `uint16_t` values to read
rc_t ats_pci_bar_read_u16(
    ats_pci_t *ats_pci, ats_pci_bar_id_t id, int offset, uint16_t *data, size_t count);

/// Writes a buffer of `uint32_t` to a BAR
///
/// @param ats_pci   PCI device to write to
/// @param space     BAR to write to
/// @param offset    Start offset in `uint32_t` in the BAR
/// @param data      Pointer to the buffer to write
/// @param count     Number of `uint32_t` values to write
rc_t ats_pci_bar_write_u32(
    ats_pci_t *ats_pci, ats_pci_bar_id_t id, int offset, uint32_t *data, size_t count);

/// Reads a buffer of uint32_t from a BAR
///
/// @param ats_pci   PCI device to write to
/// @param space     BAR to read from
/// @param offset    Start offset in `uint32_t` in the BAR
/// @param data      Pointer to the buffer to read
/// @param count     Number of `uint32_t` values to read
rc_t ats_pci_bar_read_u32(
    ats_pci_t *ats_pci, ats_pci_bar_id_t id, int offset, uint32_t *data, size_t count);

/// Writes a buffer of `uint64_t` to a BAR
///
/// @param ats_pci   PCI device to write to
/// @param space     BAR to write to
/// @param offset    Start offset in `uint64_t` in the BAR
/// @param data      Pointer to the buffer to write
/// @param count     Number of `uint64_t` values to write
rc_t ats_pci_bar_write_u64(
    ats_pci_t *ats_pci, ats_pci_bar_id_t id, int offset, uint64_t *data, size_t count);

/// Reads a buffer of uint64_t from a BAR
///
/// @param ats_pci   PCI device to write to
/// @param space     BAR to read from
/// @param offset    Start offset in `uint64_t` in the BAR
/// @param data      Pointer to the buffer to read
/// @param count     Number of `uint64_t` values to read
rc_t ats_pci_bar_read_u64(
    ats_pci_t *ats_pci, ats_pci_bar_id_t id, int offset, uint64_t *data, size_t count);

/// Writes memory buffer to a BAR
///
/// @param ats_pci   PCI device to write to
/// @param space     BAR to write to
/// @param offset    Start offset in bytes in the BAR
/// @param data      Pointer to the buffer to write
/// @param size      Size in bytes to write
rc_t ats_pci_bar_write_mem(
    ats_pci_t *ats_pci, ats_pci_bar_id_t id, size_t offset, uint8_t *data, size_t size);

/// Reads memory buffer from a BAR
///
/// @param ats_pci   PCI device to write to
/// @param space     BAR to read from
/// @param offset    Start offset in bytes in the BAR
/// @param data      Pointer to the buffer to read
/// @param size      Size in bytes to read
rc_t ats_pci_bar_read_mem(
    ats_pci_t *ats_pci, ats_pci_bar_id_t id, size_t offset, uint8_t *data, size_t size);

///
/// @brief Write an uint32_t bit field in a BAR
/// @param space     A PCI space structure to write bit field to
/// @param field     A bit field structure
/// @param value     The value to write in the bit field
rc_t ats_pci_bar_set_field(ats_pci_t *ats_pci, ats_pci_bar_regfield_t field, uint32_t value);

/// Reads a uint32_t bit field from a BAR
///
/// @param ats_pci   A PCI device to read from
/// @param space     BAR to read bit from
/// @param field     A bit field structure
/// @param pvalue    A pointer to the value to read from the bit field
rc_t ats_pci_bar_get_field(ats_pci_t *ats_pci, ats_pci_bar_regfield_t field, uint32_t *value);

/// Reads a field from a PCI BAR, and compares the result a reference value. This function returns
/// rc_t::rc_success if the values compare equal, rc_t::rc_value_mismatch if the values are
/// different, and another return code if an error happens.
///
/// @param ats_pci    A PCI device to read from
/// @param space      BAR to read from
/// @param field      A bit field structure
/// @param reference  The reference value to compare with the bit field
rc_t ats_pci_bar_field_is_equal(ats_pci_t *ats_pci,
                                ats_pci_bar_regfield_t field,
                                uint32_t reference);

/// Wait until a uint32_t bit field read from a BAR is equal to a reference value
///
/// @param ats_pci     A PCI device to wait on
/// @param space       BAR to read from
/// @param field       A bit field structure
/// @param reference   The reference value to compare with the bit field
/// @param timeout_us  The maximum delay in micro second to wait for
rc_t ats_pci_bar_wait_until_field_equals(ats_pci_t *ats_pci,
                                         ats_pci_bar_regfield_t field,
                                         uint32_t reference,
                                         uint32_t timeout_us);

/// Writes a buffer of `uint8_t` values to a PCI device's config space
///
/// @param ats_pci   PCI device to write to
/// @param offset    Start offset in `uint8_t` in the PCI device's config space
/// @param data      Pointer to the buffer to write
/// @param count     Number of `uint8_t` values to write
rc_t ats_pci_config_write_u8(ats_pci_t *ats_pci, int offset, uint8_t *data, size_t count);

/// Reads a buffer of `uint8_t` values from a PCI device's config space
///
/// @param ats_pci   PCI device to read from
/// @param offset    Start offset in `uint8_t` in the PCI device's config space
/// @param data      Pointer to the buffer to read
/// @param count     Number of `uint8_t` values to read
rc_t ats_pci_config_read_u8(ats_pci_t *ats_pci, int offset, uint8_t *data, size_t count);

/// Writes a buffer of `uint16_t` values to a PCI device's config space
///
/// @param ats_pci   PCI device to write to
/// @param offset    Start offset in `uint16_t` in the PCI device's config space
/// @param data      Pointer to the buffer to write
/// @param count     Number of `uint16_t` values to write
rc_t ats_pci_config_write_u16(ats_pci_t *ats_pci, int offset, uint16_t *data, size_t count);

/// Reads a buffer of `uint16_t` values from a PCI device's config space
///
/// @param ats_pci   PCI device to read from
/// @param offset    Start offset in `uint16_t` in the PCI device's config space
/// @param data      Pointer to the buffer to read
/// @param count     Number of `uint16_t` values to read
rc_t ats_pci_config_read_u16(ats_pci_t *ats_pci, int offset, uint16_t *data, size_t count);

/// Writes a buffer of `uint32_t` to a PCI device's config space
///
/// @param ats_pci   PCI device to write to
/// @param offset    Start offset in `uint32_t` in the PCI device's config space
/// @param data      Pointer to the buffer to write
/// @param count     Number of `uint32_t` values to write
rc_t ats_pci_config_write_u32(ats_pci_t *ats_pci, int offset, uint32_t *data, size_t count);

/// Reads a buffer of uint32_t from a PCI device's config space
///
/// @param ats_pci   PCI device to write to
/// @param offset    Start offset in `uint32_t` in the PCI device's config space
/// @param data      Pointer to the buffer to read
/// @param count     Number of `uint32_t` values to read
rc_t ats_pci_config_read_u32(ats_pci_t *ats_pci, int offset, uint32_t *data, size_t count);

/// Write an uint32_t bit field in a PCI device's config space
///
/// @param ats_pci   PCI device to write to
/// @param field     A bit field structure
/// @param value     The value to write in the bit field
rc_t ats_pci_config_set_field(ats_pci_t *ats_pci, ats_pci_config_regfield_t field, uint32_t value);

/// Reads a uint32_t bit field from a PCI device's config space
///
/// @param ats_pci   A PCI device to read from
/// @param field     A bit field structure
/// @param pvalue    A pointer to the value to read from the bit field
rc_t ats_pci_config_get_field(ats_pci_t *ats_pci, ats_pci_config_regfield_t field, uint32_t *value);

/// Reads a field from a PCI device's config space, and compares the result a reference value. This
/// function returns rc_t::rc_success if the values compare equal, rc_t::rc_value_mismatch if the
/// values are different, and another return code if an error happens.
///
/// @param ats_pci    A PCI device to read from
/// @param field      A bit field structure
/// @param reference  The reference value to compare with the bit field
rc_t ats_pci_config_field_is_equal(ats_pci_t *ats_pci,
                                   ats_pci_config_regfield_t field,
                                   uint32_t reference);

/// Wait until a uint32_t bit field read from a PCI device's config space is equal to a reference
/// value
///
/// @param ats_pci     A PCI device to wait on
/// @param field       A bit field structure
/// @param reference   The reference value to compare with the bit field
/// @param timeout_us  The maximum delay in micro second to wait for
rc_t ats_pci_config_wait_until_field_equals(ats_pci_t *ats_pci,
                                            ats_pci_config_regfield_t field,
                                            uint32_t reference,
                                            uint32_t timeout_us);

/// Retreive PCI common config space
///
/// @param ats_pci       A PCI device
/// @param config        A pointer to the common config structure used to store the result
rc_t ats_pci_common_config(ats_pci_t *ats_pci, ats_pci_common_config_t *config);

/// Retreive PCI express capabilities
///
/// @param ats_pci       A PCI device
/// @param capabilities  A pointer to the capabilities structure used to store the result
rc_t ats_pci_express_capabilities(ats_pci_t *ats_pci, ats_pci_express_capabilities_t *capabilities);

/// Initialize `ats_pci` structure with informations founds in the PCI config space
///
/// @param ats_pci       A PCI device
/// @param config_read   A pointer to function used to read from the PCI config space
/// @param config_write  A pointer to function used to write to the PCI config space
/// @param driver        A pointer to the OS driver context
rc_t ats_pci_init(ats_pci_t *ats_pci,
                  ats_pci_config_read_f config_read,
                  ats_pci_config_write_f config_write,
                  driver_os_handle_t *driver);

/// Register a PCI interrupt request.
///
/// @param ats_pci       PCI device to register interrupt to
/// @param irq           Interrupt request Id number
/// @param handler       Function called by the operating system when interrupt occurs
/// @param enabler       Function called by the operating system to enable or disable interrupt
/// @param context       Callbacks context pointer.
rc_t ats_pci_register_irq(ats_pci_t *ats_pci,
                          int irq,
                          interrupt_handler_t handler,
                          on_interrupt_enable_request_t on_enable_request,
                          void *context);

/// Enable or disable a PCI interrupt request.
///
/// @param ats_pci       PCI device to register interrupt to
/// @param irq           Interrupt request Id number
rc_t ats_pci_enable_irq(ats_pci_t *ats_pci, int irq);

/// Enable or disable a PCI interrupt request.
///
/// @param ats_pci       PCI device to register interrupt to
/// @param irq           Interrupt request Id number
rc_t ats_pci_disable_irq(ats_pci_t *ats_pci, int irq);

static inline ats_pci_bar_t *ats_pci_get_bar(ats_pci_t *ats_pci, ats_pci_bar_id_t bar_id)
{
    return (((size_t)bar_id < (size_t)ATS_PCI_BAR_MAX) ? &ats_pci->bars[bar_id] : NULL);
}

#endif // ATS_PCI_H
