
#include "ats_procfs.h"
#include "ats_pci.h"
#include "utils.h"

#include "driver_version.h"

static int show_pci_config(void *context, struct seq_file *sfile)
{
    ats_pci_t *ats_pci = (ats_pci_t *)context;
    ats_pci_common_config_t config;

    if (ats_pci_common_config(ats_pci, &config) != rc_success) return 0;

    at_seq_printf(sfile, "PCI Configuration:\n");

    at_seq_printf(sfile, "  VendorID:                     0x%04X\n", config.vendor_id);
    at_seq_printf(sfile, "  DeviceID:                     0x%04X\n", config.device_id);
    at_seq_printf(sfile, "  Command:                      0x%04X\n", config.command);
    at_seq_printf(sfile, "  Status:                       0x%04X\n", config.status);
    at_seq_printf(sfile, "  RevisionID:                   0x%02X\n", config.revision_id);
    at_seq_printf(sfile, "  ProgIf:                       0x%02X\n", config.prog_if);
    at_seq_printf(sfile, "  SubClass:                     0x%02X\n", config.sub_class);
    at_seq_printf(sfile, "  BaseClass:                    0x%02X\n", config.base_class);
    at_seq_printf(sfile, "  CacheLineSize:                0x%02X\n", config.cache_line_size);
    at_seq_printf(sfile, "  LatencyTimer:                 0x%02X\n", config.latency_timer);
    at_seq_printf(sfile, "  HeaderType:                   0x%02X\n", config.header_type);
    at_seq_printf(sfile, "  BuiltInSelfTest:              0x%02X\n", config.BIST);

    for (size_t id = 0; id < ARRAY_SIZE(config.raw_base_addresses); id++)
        at_seq_printf(sfile, "  BAR[%d]:                       0x%08X\n", id,
                      config.raw_base_addresses[id]);

    at_seq_printf(sfile, "  CardBusCISPtr:                0x%08X\n", config.card_bus_CIS_pointer);
    at_seq_printf(sfile, "  SubsystemVendorID:            0x%04X\n", config.subsystem_vendor_id);
    at_seq_printf(sfile, "  SubsystemID:                  0x%04X\n", config.subsystem_id);
    at_seq_printf(sfile, "  ROMBaseAddress:               0x%08X\n", config.ROM_base_address);
    at_seq_printf(sfile, "  CapabilitiesPtr:              0x%02X\n", config.capabilities_pointer);
    at_seq_printf(sfile, "  InterruptLine:                0x%02X\n", config.interrupt_line);
    at_seq_printf(sfile, "  InterruptPin:                 0x%02X\n", config.interrupt_pin);
    at_seq_printf(sfile, "  MinimumGrant:                 0x%02X\n", config.minimum_grant);
    at_seq_printf(sfile, "  MaximumLatency:               0x%02X\n", config.maximum_latency);

    size_t offset = OFFSET_OF(ats_pci_common_config_t, device_specific);

    at_seq_printf(sfile, "  DeviceSpecific (offset):      0x%03x\n", offset);

    for (size_t id = 0; id < ARRAY_SIZE(config.device_specific); id++)
    {
        if ((id % 16) == 0) at_seq_printf(sfile, "\noffset 0x%03x: ", offset + id);

        at_seq_printf(sfile, " %02X", config.device_specific[id]);
    }

    at_seq_printf(sfile, "\n");
    return 0;
}

static int show_card_config(void *context, struct seq_file *sfile)
{
    board_config_database_t *database = (board_config_database_t *)context;
    const ALAZAR_BOARD_CONFIG_ID *pid;

    for (pid = database->def.ids; *pid != ALAZAR_BOARD_CONFIG_NONE; pid++)
    {
        ALAZAR_BOARD_CONFIG_VALUE value;

        ATS_CHECK(board_config_get(database, *pid, &value));

        switch (*pid)
        {
            case ALAZAR_BOARD_CONFIG_BUS_TYPE:
                at_seq_printf(sfile, "bus_type:             %u\n", value.vu16);
                break;

            case ALAZAR_BOARD_CONFIG_BOARD_TYPE:
                at_seq_printf(sfile, "board_type:           %u\n", value.vu32);
                break;

            case ALAZAR_BOARD_CONFIG_SERIAL_NUMBER:
                at_seq_printf(sfile, "serial_number:        %u\n", value.vu32);
                break;

            case ALAZAR_BOARD_CONFIG_MFG_BATCH:
                at_seq_printf(sfile, "mfg_batch:            %u\n", value.vu32);
                break;

            case ALAZAR_BOARD_CONFIG_TESTED_BY:
                at_seq_printf(sfile, "tested_by:            %u\n", value.vu32);
                break;

            case ALAZAR_BOARD_CONFIG_ON_BOARD_MEMORY:
                at_seq_printf(sfile, "on_board_memory:      %u\n", value.vu8);
                break;

            case ALAZAR_BOARD_CONFIG_OPTIONS:
                at_seq_printf(sfile, "options:              0x%016llx\n", value.vu64);
                break;

            case ALAZAR_BOARD_CONFIG_FIRST_CAL_DATE:
                at_seq_printf(sfile, "first_cal_date:       %04d/%02d/%02d\n", value.vdate.year,
                              value.vdate.month, value.vdate.day);
                break;

            case ALAZAR_BOARD_CONFIG_LAST_CAL_DATE:
                at_seq_printf(sfile, "last_cal_date:        %04d/%02d/%02d\n", value.vdate.year,
                              value.vdate.month, value.vdate.day);
                break;

            case ALAZAR_BOARD_CONFIG_NEXT_CAL_DATE:
                at_seq_printf(sfile, "next_cal_date:        %04d/%02d/%02d\n", value.vdate.year,
                              value.vdate.month, value.vdate.day);
                break;

            case ALAZAR_BOARD_CONFIG_TEST_DATE:
                at_seq_printf(sfile, "test_date:            %04d/%02d/%02d\n", value.vdate.year,
                              value.vdate.month, value.vdate.day);
                break;

            default:
                break;
        }
    }

    return 0;
}

static int show_driver_info(void *context, struct seq_file *sfile)
{
    at_seq_printf(sfile, "version:  %d.%d.%d\n", DRIVER_VERSION_MAJOR, DRIVER_VERSION_MINOR,
                  DRIVER_VERSION_PATCH);
    return 0;
}

static int show_fpga_info(void *context, struct seq_file *sfile)
{
    ats_dev_t *ats_dev = (ats_dev_t *)context;

    for (int reg = 0; reg < 64; reg++)
    {
        uint32_t value = 0;

        ATS_CHECK(ats_dev_reg_space_register_read(ats_dev, ats_dev_reg_space_fpga, reg, &value));
        at_seq_printf(sfile, "register[%02d] 0x%08x\n", reg, value);
    }

    return 0;
}

typedef struct
{
    void *context;
    ats_procfs_show_func_t show;
    ats_procfs_write_func_t write;
} ats_procfs_entry_t;

int at_procfs_show(struct seq_file *sfile, void *offset)
{
    ats_procfs_entry_t *entry = (ats_procfs_entry_t *)at_seq_get_private(sfile);

    if (entry == NULL)
    {
        ats_log_error("Invalid procfs entry\n");
        return -1;
    }

    if (entry->show == NULL)
    {
        ats_log_error("No procfs show function\n");
        return -1;
    }

    return (entry->show)(entry->context, sfile);
}

ssize_t at_procfs_write(struct file *file, const char __user *buffer, size_t length, loff_t *offset)
{
    struct seq_file *sfile = (struct seq_file *)at_file_get_private_data(file);
    ats_procfs_entry_t *entry = (ats_procfs_entry_t *)at_seq_get_private(sfile);

    if (entry == NULL)
    {
        ats_log_error("Invalid procfs entry\n");
        return -1;
    }

    if (entry->write == NULL)
    {

        ats_log_error("No procfs write function\n");
        return -1;
    }

    ssize_t count;

    if ((count = (entry->write)(entry->context, buffer, length, offset)) > 0) *offset += count;

    return count;
}

rc_t ats_procfs_create(ats_procfs_t *procfs, const char *dir)
{
    return (((procfs->rootdir = at_proc_mkdir(dir, NULL)) == NULL) ? rc_failed : rc_success);
}

rc_t ats_procfs_destroy(ats_procfs_t *procfs)
{
    if (procfs->rootdir == NULL) return rc_failed;

    at_proc_remove(procfs->rootdir);
    return rc_success;
}

rc_t ats_procfs_add(ats_procfs_t *procfs,
                    const char *name,
                    mode_t mode,
                    ats_procfs_show_func_t show,
                    ats_procfs_write_func_t write,
                    void *context)
{
    ats_procfs_entry_t *entry;

    if ((entry = (ats_procfs_entry_t *)ats_alloc(sizeof(*entry))) == NULL)
    {
        ats_log_error("Unable to allocate procfs entry for '%s'\n", name);
        return rc_failed;
    }

    memset(entry, 0, sizeof(*entry));

    struct proc_dir_entry *dirent = at_procfs_create_data(name, mode, procfs->rootdir, entry);

    if (dirent == NULL)
    {
        ats_log_error("Unable to create procfs entry for '%s'\n", name);
        return rc_failed;
    }

    entry->context = context;
    entry->show = show;
    entry->write = write;
    return rc_success;
}

rc_t ats_procfs_attach(ats_procfs_t *procfs,
                       const char *dir,
                       ats_pci_t *pci,
                       board_config_database_t *config,
                       ats_dev_t *ats_dev)
{
    ATS_CHECK(ats_procfs_create(procfs, dir));
    ATS_CHECK(ats_procfs_add(procfs, "pci", S_IFREG | S_IRUGO, show_pci_config, NULL, pci));
    ATS_CHECK(ats_procfs_add(procfs, "card", S_IFREG | S_IRUGO, show_card_config, NULL, config));
    ATS_CHECK(ats_procfs_add(procfs, "driver", S_IFREG | S_IRUGO, show_driver_info, NULL, NULL));
    ATS_CHECK(ats_procfs_add(procfs, "fpga", S_IFREG | S_IRUGO, show_fpga_info, NULL, ats_dev));
    return rc_success;
}

rc_t ats_procfs_detach(ats_procfs_t *procfs) { return ats_procfs_destroy(procfs); }
