/**
 *  @file platform.h
 *
 *  Collection of platform-agnostic utility functions.
 *
 */

#ifndef PLATFORM_H
#define PLATFORM_H

// Includes
/////////////////////////////////////////////////////////////////////////////

#if defined(_DRIVER_) && defined(_LINUX_)
#ifdef MODULE
// ===== Linux kernel ===============================================
#include <linux/kernel.h>
#else
// ===== linux/common ===============================================
#include <kernel_types.h>
#endif
#endif

// ===== src ================================================================
#include "ats_types.h"

#ifdef __cplusplus
#define EXTERN_C_BEGIN                                                                             \
    extern "C"                                                                                     \
    {
#define EXTERN_C_END }
#else // __cplusplus
#define EXTERN_C_BEGIN
#define EXTERN_C_END
#endif // __cplusplus

EXTERN_C_BEGIN

#ifdef _WIN32

#pragma warning(disable : 4100 4204 4214)

#define IOCTL_USER_POINTER void *

#define ats_strlen wcslen
#define ats_strnlen wcsnlen
#define ats_strncpy wcsncpy
#define ats_strcmp wcscmp
#define ats_strncmp wcsncmp
#define ats_snprintf sprintf_s

#ifdef _DRIVER_

#include <Ntddk.h>
#include <Ntstrsafe.h>

void ats_log_init(PDEVICE_OBJECT pobj);

#else // _DRIVER_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void ats_log_init(uint32_t mask);
uint32_t ats_log_mask(void);

#endif // _DRIVER_

/// Calls the expression `x`, which must return a NTSTATUS. If the return value
/// is not a success, logs an error with ats_log_error and returns the
/// erroneous status code.
#define ATS_NT_RCHECK(x)                                                                           \
    do                                                                                             \
    {                                                                                              \
        NTSTATUS stat = (x);                                                                       \
        if (!NT_SUCCESS(stat))                                                                     \
        {                                                                                          \
            ats_log_error("<-- Error in %s(): 0x%x\n", __FUNCTION__, stat);                        \
            return stat;                                                                           \
        }                                                                                          \
    } while (0)

#else // _WIN32

#define IOCTL_USER_POINTER void __user *

#define ats_strlen strlen
#define ats_strnlen strnlen
#define ats_strncpy strncpy
#define ats_strcmp strcmp
#define ats_strncmp strncmp
#define ats_snprintf snprintf

#ifdef _DRIVER_

#include <linux/string.h>

extern int snprintf(char *, size_t, const char *, ...);

#else // _DRIVER_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#endif // _DRIVER_

void ats_log_init(uint32_t mask);
uint32_t ats_log_mask(void);

#endif // _WIN32

enum ATS_LOG_LEVEL
{
    ATS_LOG_ERROR,
    ATS_LOG_WARNING,
    ATS_LOG_INFO,
    ATS_LOG_DEBUG,
    ATS_LOG_TEXT,
    ATS_LOG_TRACE,
    ATS_LOG_DEBUG_LVL1 = 0x10,
    ATS_LOG_DEBUG_LVL2,
    ATS_LOG_DEBUG_LVL3,
};

#define ATS_MASK(__o) (1UL << (__o))
#define ATS_LOG_MASK_DEFAULT                                                                       \
    (ATS_MASK(ATS_LOG_ERROR) | ATS_MASK(ATS_LOG_WARNING) | ATS_MASK(ATS_LOG_TEXT))

#define ats_log_error(__f, ...) ats_log(ATS_LOG_ERROR, __f, ##__VA_ARGS__)
#define ats_log_warning(__f, ...) ats_log(ATS_LOG_WARNING, __f, ##__VA_ARGS__)
#define ats_log_info(__f, ...) ats_log(ATS_LOG_INFO, __f, ##__VA_ARGS__)
#define ats_log_debug(__f, ...) ats_log(ATS_LOG_DEBUG, __f, ##__VA_ARGS__)
#define ats_log_text(__f, ...) ats_log(ATS_LOG_TEXT, __f, ##__VA_ARGS__)
#define ats_log_trace(__f, ...) ats_log(ATS_LOG_TRACE, __f, ##__VA_ARGS__)

void ats_log(int level, const char *format, ...);

void ats_write_register_u32(uint32_t *pmem, uint32_t val);
uint32_t ats_read_register_u32(uint32_t *pmem);

void ats_udelay(unsigned long delay);
void ats_mdelay(unsigned long delay);
int64_t ats_get_utime(void);

/// On Linux allocated memory pages are guaranteed to be contiguous (Size is limited to 128 KBytes).
/// On Windows allocated memory pages are not guaranteed to be contiguous.
void *ats_alloc(size_t size);
void ats_free(void *buffer);

/// Allocated memory pages are not guaranteed to be contiguous.
void *ats_alloc_large_buffer(size_t size);
void ats_free_large_buffer(void *buffer);

/// Copy `size_bytes` of data from the kernel buffer `src` to the user buffer
/// `dst`. Returns 0 upon success
int ats_copy_to_user(void *dst, void *src, size_t size_bytes);

/// Copy `size_bytes` of data from the user buffer `src` to the kernel buffer
/// `dst`. Returns 0 upon success
int ats_copy_from_user(void *dst, void *src, size_t size_bytes);

typedef uint64_t ioctl_ptr_t;

#ifdef _DRIVER_

static inline IOCTL_USER_POINTER ioctl_user_pointer(bool io32, ioctl_ptr_t ptr)
{
    if ((sizeof(IOCTL_USER_POINTER) < 8) || (io32)) ptr &= ((1ULL << 32) - 1);
    return (IOCTL_USER_POINTER)((sizeof(IOCTL_USER_POINTER) < 8) ? (uint32_t)ptr : ptr);
}

static inline int ioctl_copy_to_user(bool io32, ioctl_ptr_t dst, void *src, size_t size)
{
    return ats_copy_to_user(ioctl_user_pointer(io32, dst), src, size);
}

static inline int ioctl_copy_from_user(bool io32, void *dst, ioctl_ptr_t src, size_t size)
{
    return ats_copy_from_user(dst, ioctl_user_pointer(io32, src), size);
}

#endif // _DRIVER_

/// Checks if a user space pointer is valid
///
/// @param addr  User space pointer to start of block to check
/// @param size  Size of block to check
bool ats_access_ok(void *addr, size_t size);

#ifndef __FUNCTION_NAME__
#define __FUNCTION_NAME__ __FUNCTION__
#endif

EXTERN_C_END

#if defined(_M_X64) || defined(__x86_64__) || defined(_AMD64_) || defined(__amd64__) ||            \
    defined(_M_ARM64) || defined(__aarch64__)
#define USE_CPU_64BITS
#endif

#endif // PLATFORM_H
